/** \brief Generates the layout for either wrapped or non-wrapped text and stores the result

Use this class for all your text output needs. It takes text with formatting
markup as input and turns that into the glyphs and their necessary positions.
It stores the glyphs internally, but maintains enough information to both
retrieve your own rendering information if you wish and to perform visual
text editing where the output refers back to where it came from.

Usage:
-# Construct
-# Set the text using appendText() and appendControlCode()
-# If you want text wrapping, set up #wrap_inside_shapes
-# Call calculateFlow()
-# You can go several directions from here, but the most interesting
   things start with creating a TextFlow::iterator with begin() or end().

Terminology, in descending order of size:
- Flow: Not often used, but when it is it means all the text
- Shape: A Shape object which is used to represent one of the regions inside
  which to flow the text. Can overlap with...
- Paragraph: Err...A paragraph. Contains one or more...
- Line: An entire horizontal line with a common baseline. Contains zero or
  more...
- Chunk: These only happen when a shape is sufficiently complex that the
  text has to flow either side of some obstruction in the middle. A chunk
  is the base unit for wrapping. Contain one or more...
- Span: A convenient subset of a chunk with the same font, style,
  directionality, block progression and input stream. Fill and outline
  need not be constant because that's a later rendering stage.
- This is where it gets weird because a span will contain one or more 
  elements of both of the following, which can overlap with each other in
  any way:
  - Character: a single Unicode codepoint from an input stream. Many arabic
    characters contain multiple glyphs
  - Glyph: a rendering primitive for font engines. A ligature glyph will
    represent multiple characters.

Other terminology:
- Input stream: An object representing a single call to appendText() or
  appendControlCode().
- Control code: Metadata in the text stream to signify items that occupy
  real space (unlike style changes) but don't belong in the text string.
  Paragraph breaks are in this category. See TextFlow::TextControlCode.
- SVG1.1: The W3C Recommendation "Scalable Vector Graphics (SVG) 1.1"
  http://www.w3.org/TR/SVG11/
- 'left', 'down', etc: These terms are generally used to mean what they
  mean in left-to-right, top-to-bottom text but rotated or reflected for
  the current directionality. Thus, the 'width' of a ttb line is actually
  its height, and the (internally stored) y coordinate of a glyph is
  actually its x coordinate. Confusing to the reader but much simpler in
  the code. All public methods use real x and y.

Comments:
- There's a strong emphasis on international support in this class, but
  that's primarily because once you can display all the insane things
  required by various languages, simple things like styling text are
  trivial.
- There are a few places (appendText() is one) where pointers are held to
  caller-owned objects and used for quite a long time. This is messy but
  is safe for our usage scenario and in many cases the cost of copying the
  objects is quite high.
- "Why isn't foo here?": Ask yourself if it's possible to implement foo
  externally using iterators. However this may not mean that it doesn't
  belong as a member, though.
- I've used floats rather than doubles to store relative distances in some
  places where it would save significant amounts of memory. The SVG spec
  allows you to do this as long as intermediate calculations are done
  double. Is this safe? Very very long lines might not finish precisely
  where you want, but that's to be expected with any typesetting.
- If you look at the six arrays for holding the output data you'll realise
  that there's no O(1) way to drill down from a paragraph to find its
  starting glyph. This was a conscious decision to reduce complexity and
  to save memory. Drilling down isn't actually that slow because you can
  go step-by-step and will rarely end up scanning more than about 50 items.
  I think I make it O(log(log(n)). Add this to the realisation that most of
  the times you do this will be in response to user actions and hence you
  only need to be faster than the user and I think the design makes sense.
- There are a massive number of functions acting on TextFlow::iterator. A
  large number are trivial and will be inline, but is it really necessary
  to have all these, especially when some can be implemented by the caller
  using the others?
- The separation of methods between TextFlow and TextFlow::iterator is a
  bit arbitrary, because many methods could go in either. I've used the STL
  model where the iterator itself can only move around; the base class is
  required to do anything interesting.
*/
class TextFlow {
public:
    class iterator;
    friend iterator;

    enum Direction {LEFT_TO_RIGHT, RIGHT_TO_LEFT, TOP_TO_BOTTOM, BOTTOM_TO_TOP};

    TextFlow();
    ~TextFlow();

    /* ************************** describing the stuff to flow *************************/

    /** Empties everything stored in this class and resets it to its
    original state, like when it was created. All iterators on this
    object will be invalidated. */
    void clear();

    /** Takes an input string of the form used for the \a x attribute (and
    others) or the SVG text tag (see SVG1.1 section 10.4) and produces a
    numeric output of it. */
    static std::vector<SPILength> parse_svg_length_list(char const *text, bool allow_units = true);

    /** The optional attributes which can be applied to a SVG text
    or related tag. See SVG1.1 section 10.4 for the definitions of all
    these members. */
    struct OptionalTextTagAttrs {
        enum LengthAdjust {SPACING, SPACING_AND_GLYPHS};
        std::vector<SPILength> x;
        std::vector<SPILength> y;
        std::vector<SPILength> dx;
        std::vector<SPILength> dy;
        std::vector<SPILength> rotate;
        SPILength text_length;        /// not yet implemented
        LengthAdjust length_adjust;   /// not yet implemented
        bool preserve_space;   /// the xml:space attribute, see SVG1,1 section 10.15
    };

    /** adds a new piece of text to the end of the current list of text to
    be processed. This method can only add text of a consistent style.
    To add lots of different styles, call it lots of times.
    Any special characters (linefeed, etc) in \a text will be interpreted
    as per the XML spec and silently ignored. Tabs are also treated as in
    XML and interpreted the same as a space character.
     \param text  The text. \b Note: only a \em pointer is stored. Do not
                  mess with the text until after you have called
                  calculateFlow().
     \param style The font style. \b Note: only a \em pointer is stored.
                  You can, however, mess with this as much as you like
                  (from the same thread) as long as you don't delete it.
                  Changes to colours and suchlike will take place as soon
                  as you next call show(), but changes to fonts and such
                  will require calling calculateFlow() again.
     \param source_cookie  This pointer is treated as opaque by TextFlow
                  but will be passed through the flowing process intact so
                  that callers can use it to refer to the original object
                  that generated a particular glyph. See TextFlow::iterator.
     \param optional_attributes  A structure containing additional options
                  for this text. See OptionalTextTagAttrs. The values are
                  copied to internal storage before this method returns.
     \param text_from_byte  Used for selecting only a substring of \a text
                  to process. This specifies the byte offset of the first
                  character to use.
     \param text_to_byte  Used for selecting only a substring of \a text
                  to process. This specifies the byte offset of the last
                  character to use. If -1, the endpoint is the end of
                  \a text.
    */
    void appendText(Glib::ustring const &text, SPStyle const &style, void *source_cookie, OptionalTextTagAttrs const *optional_attributes = NULL, int text_from_byte = 0, int text_to_byte = -1);

    enum TextControlCode {
        PARAGRAPH_BREAK,    /// forces the flow to move on to the next line
        SHAPE_BREAK,        /// forces the flow to ignore the remainder of the current shape (from #flow_inside_shapes) and continue at the top of the one after.
        ARBITRARY_GAP       /// inserts an arbitrarily-sized hole in the flow in line with the current text.
    };
    /** Control codes are metadata in the text stream to signify items
    that occupy real space (unlike style changes) but don't belong in the
    text string. See TextControlCode for the types available.
     \param code    A member of the TextFlowControlCode enumeration.
     \param width   The width in pixels that this item occupies.
     \param ascent  The number of pixels above the text baseline that this
                    control code occupies.
     \param descent The number of pixels below the text baseline that this
                    control code occupies.
     \param source_cookie This pointer is treated as opaque by TextFlow
                  but will be passed through the flowing process intact so
                  that callers can use it to refer to the original object
                  that generated a particular area. See TextFlow::iterator.
    Note that for some control codes (eg tab) the values of the \a width,
    \a ascender and \a descender are implied by the surrounding text (and
    in the case of tabs, the values set in tab_stops) so the values you pass
    here are ignored.
    */
    void appendControlCode(TextControlCode code, double width = 0.0, double ascent = 0.0, double descent = 0.0, void *source_cookie = NULL);

    struct WrapShape {
        enum DisplayAlign {ALIGN_BEFORE, ALIGN_CENTER, ALIGN_AFTER};
        Shape *shape;
        DisplayAlign display_align;    /// see XSL1.0 section 7.13.4. The behaviour of settings other than ALIGN_BEFORE for non-rectangular shapes is undefined.
    };
    /** Stores the shapes inside which to flow the text. If this array is
    empty then no automatic wrapping is done and lines will continue to
    infinity if necessary. Text can be flowed inside multiple shapes in
    sequence, like with frames in a DTP package. If the text flows past
    the end of the last shape all remaining text is ignored.

    The storage for the Shape objects is managed by the caller, and need
    only remain valid for the duration of the call to calculateFlow().
    */
    std::vector<WrapShape> wrap_inside_shapes;

    /* ************************** doing the actual flowing *************************/

    /** Takes all the stuff you set with the members above here and creates
    a load of glyphs for use with the members below here. All iterators on
    this object will be invalidated.
      \return  false on failure.
    */
    bool calculateFlow();

    /* ************************** operating on the output glyphs *************************/
    // (see also TextFlow::iterator) */

    /** Adds all the output glyphs to \a in_arena using the given \a paintbox.
     \param in_arena  The arena to add the glyphs group to
     \param paintbox  The current rendering tile
    */
    void show(NRArenaGroup *in_arena, NRRect const *paintbox) const;

    /** Calculates the smallest rectangle completely enclosing all the
    glyphs.
      \param bounding_box  Where to store the box
      \param transform     The transform to be applied to the entire object
                           prior to calculating its bounds.
    */
    void getBoundingBox(NRRect *bounding_box, const NR::Matrix &transform) const;

    /** Sends all the glyphs to the given print context.
     \param ctx   I have
     \param pbox  no idea
     \param dbox  what these
     \param bbox  parameters
     \param ctm   do yet
    */
    void print(SPPrintContext *ctx, NRRect const *pbox, NRRect const *dbox, NRRect const *bbox, NRMatrix const &ctm) const;

    /** debug and unit test method. Creates a textual representation of the
    contents of this object. The output is designed to be both human-readable
    and comprehensible when diffed with a known-good dump. */
    Glib::ustring dumpAsText() const;

    /** Moves all the glyphs in the structure so that the baseline of all
    the characters sits neatly along the path specified. If the text has
    more than one line the results are undefined. The 'align' means to
    use the SVG align method as documented in SVG1.1 section 10.13.2. */
    void fitToPathAlign(SPILength const &startOffset,Path const &path);

    /** Convert the specified range of characters into their bezier 
    outlines.
    */
    SPCurve* convertToCurves(iterator const &from_glyph, iterator const &to_glyph) const;
    SPCurve* convertToCurves() const
        {return convertToCurves(begin(), end());}

    /** Apply the given transform to all the output presently stored in
    this object. */
    void transform(const NR::Matrix &transform);

    //******* iterator methods

    /** Returns an iterator pointing at the first glyph of the flowed output.
    The first glyph is also the first character, line, paragraph, etc. */
    iterator begin() const;

    /** Returns an iterator pointing just past the end of the last glyph,
    which is also just past the end of the last chunk, span, etc, etc. */
    iterator end() const;

    /** Returns an iterator pointing to the cursor position for a mouse
    click at the given coordinates. */
    iterator getNearestCursorPositionTo(double x, double y) const;
    iterator getNearestCursorPositionTo(NR::Point &point) const
        {return getNearestCursorPositionTo(point[0], point[1]);}

    /** Returns an iterator pointing to the letter whose bounding box contains
    the given coordinates. end() if the point is not over any letter. */
    iterator getLetterAt(double x, double y) const;
    iterator getLetterAt(NR::Point &point) const
        {return getLetterAt(point[0], point[1]);}

    /** Returns an iterator pointing to the character in the output which
    was created from the given input. If the character at the given byte
    offset was removed (due to OptionalTextTagAttrs::preserve_space, for
    example) the next character after it is returned. \a byte_offset is
    relative to the beginning of the text passed to appendText(), ie it
    is greater than the \a text_from_byte parameter. If the parameters
    are out of bounds, end() is returned. */
    iterator sourceToGlyph(void *source_cookie, int byte_offset) const;

    // many functions acting on iterators, most of white are obvious

    int glyphAt(iterator const &it) const;   /// the actual glyph, not its index
    void glyphBoundingBox(iterator const &it, NR::Rect *box, double *rotation) const;
    void chunkBoundingBox(iterator const &it, NR::Rect *box) const;
    int lineIndex(iterator const &it) const;
    void lineBoundingBox(iterator const &it, NR::Rect *box) const;
    int shapeIndex(iterator const &it) const;
    bool isWhitespace(iterator const &it) const;
    int characterAt(iterator const &it) const;   /// the actual character, not its index
    void getSourceOfCharacter(iterator const &it, void **source_cookie, int *byte_offset) const;
    void characterBoundingBox(iterator const &it, NR::Rect *box, double *rotation) const;
    /** Basically uses character_bounding_box() on all the characters between
    \a start and \a end and returns the union of these boxes. */
    Shape* createSelectionShape(iterator const &it_start, iterator const &it_end) const;
    bool isCursorPosition(iterator const &it) const;
    /** Gets the ideal cursor shape for a given iterator. The result is
    undefined if \a it is not at a valid cursor position.
      \param it        The location in the output
      \param position  The pixel location of the centre of the 'bottom' of
                       the cursor.
      \param height    The height in pixels of the surrounding text
      \param rotation  The angle to draw from \a position. Degrees, zero up,
                       increasing clockwise.
      \param skew      The skew angle to use. Nonzero when the surrounding
                       text is oblique or italic.
    */
    void queryCursorShape(iterator const &it, NR::Point *position, double *height, double *rotation, double *skew) const;
    bool isStartOfWord(iterator const &it) const;
    bool isEndOfWord(iterator const &it) const;
    bool isStartOfSentence(iterator const &it) const;
    bool isEndOfSentence(iterator const &it) const;
    int paragraphIndex(iterator const &it) const;

private:
    /** used for specifying anything which is a range. The units are
    unspecified, but usually bytes. */
    struct TextRange {
        int from, to;
    };

    /** Represents a text item in the input stream. See #_input_stream.
    Most of the members are copies of the values passed to appendText(). */
    struct InputStreamTextSource {
        struct TextBytesRemovalEntry {
            int number_of_used_bytes;
            int number_of_omitted_bytes;
        };
        Glib::ustring const *text;
        TextRange text_range_bytes;
        SPStyle const *style;
        void *source_cookie;
        /** Note that in the SVG spec (1.1 section 10.5) the length vectors
        in here are cumulative. By the time the data has been put here,
        however, appendText() has already resolved the overlaps and these
        numbers are absolute. */
        OptionalTextTagAttrs optional_attributes;
        /** In order to be able to translate a character index in the output
        into a character index in the input sources, even in the presence of
        preserve_space == false, this member keeps track of the bytes from
        this source that were never used anywhere in the rendering process. */
        std::vector<TextBytesRemovalEntry> text_bytes_removed;
    };

    /** Represents a control code item in the input stream. See
    #_input_streams. All the members are copies of the values passed to
    appendControlCode(). */
    struct InputStreamControlCode {
        TextControlCode code;
        double ascent, descent, width;
        void *source_cookie;
    };

    enum InputStreamItemType {TEXT_SOURCE, CONTROL_CODE};
    /** Collects the different types of items in the stream into one structure
    for easier processing. */
    struct InputStreamItem {
        InputStreamItemType type;
        union {
            InputStreamTextSource text_source;
            InputSreamControlCode control_code;
        };
    };
    /** This is our internal storage for all the stuff passed to the
    appendText() and appendControlCode() functions. */
    std::vector<InputStreamItem> _input_stream;

    struct InputWrapShapeData {
        /** For vertical text it will be necessary to rotate the input Shape
        by 90 degrees so that the scanline processing routines can be used
        to generate the line widths. This stores the rotated copy. It would
        be more efficient to have additional scanline routines that can go
        sideways, but it would also be more work. */
        Shape *rotated_shape;
    };
    std::vector<InputWrapShapeData> _input_wrap_shapes;

    /** This class from the old routines is decently self-contained and
    can be used to generate scanlines of the input shapes. */
    flow_dest flow_dst;

    struct Glyph {
        int glyph;
        int in_character;
        float cluster_x; /// different from #x for, eg, cursive 'f' where part of the glyph precedes the cursor position. Relative to the start of the line
        float x;         /// relative to the start of the line
        float y;         /// relative to the current line's baseline
        float rotation;  /// absolute, modulo any object transforms we don't know about
        // in_span not required: use character[in_character].in_span
    };
    struct Character {
        int in_span;
        PangoLogAttr pangoAttributes;
        int in_glyph;
    };
    struct Span {
        int in_chunk;
        font_instance *font;
        float font_size;
        float ascent, descent;
        Direction direction;     /// See CSS3 section 3.2. Either rtl or ltr
        Direction block_progression;  /// See CSS3 section 3.2. The direction in which lines go.
        int in_input_stream_item;
        /** many spans will only contain part of the text initially set in
        an input stream. This records the byte index of the first character
        of the span in the stream #in_input_stream_item. That's the byte
        offset \em after whitespace has been sorted, so you've still got to
        use InputStreamTextSource::text_bytes_removed. */
        int offset_of_first_char_in_input_stream;
    };
    struct Chunk {
        int in_line;
        double left_x;
    };
    struct Line {
        int in_paragraph;
        double baseline_y;
        int in_shape;
    };
    struct Paragraph {
        Direction base_direction;
    };
    std::vector<Paragraph> _paragraphs;
    std::vector<Line> _lines;
    std::vector<Chunk> _chunks;
    std::vector<Span> _spans;
    std::vector<Character> _characters;
    std::vector<Glyph> _glyphs;
}

/** \brief Holds a position within the glyph output of TextFlow.

Used to access the output of a TextFlow, query information and generally
move around in it. See TextFlow for a glossary of the names of functions.

There is no public constructor for this class. Use one of the methods from
TextFlow to create an instance.

I'm not going to document all the methods because most of their names make
their function self-evident.

The next_ and prev_* methods all return false if the requested action would
have caused the current position to move out of bounds. In this case no
change is made.
*/
class TextFlow::iterator {
    // no copy constructor required, the default does what we want
    bool operator== (iterator const &other);
    bool operator!= (iterator const &other);
    bool operator< (iterator const &other);
    bool operator<= (iterator const &other);
    bool operator> (iterator const &other);
    bool operator>= (iterator const &other);

    /* **** visual-oriented methods **** */

    //glyphs
    bool nextGlyph();
    bool prevGlyph();

    //chunk
    bool nextStartOfChunk();
    bool prevStartOfChunk();

    //line
    bool nextStartOfLine();
    bool prevStartOfLine();

    //shape
    bool nextStartOfShape();
    bool prevStartOfShape();

    /* **** text-oriented methods **** */

    //characters
    bool nextCharacter();
    bool prevCharacter();

    bool nextCursorPosition();
    bool prevCursorPosition();
    void nextLineCursor();
    void prevLineCursor();
    void beginCursorUpDown();  /// stores the current x coordinate so that the cursor won't drift. See #_x_coordinate

    //words
    bool nextStartOfWord();
    bool prevStartOfWord();
    bool nextEndOfWord();
    bool prevEndOfWord();

    //sentences
    bool nextStartOfSentence();
    bool prevStartOfSentence();
    bool nextEndOfSentence();
    bool prevEndOfSentence();

    //paragraphs
    bool nextStartOfParagraph();
    bool prevStartOfParagraph();
    //no end_of_para methods because that's just the previous char

    //sources
    bool nextStartOfSource();
    bool prevStartOfSource();

private:
    TextFlow const *_parent_textflow;
    int _current_glyph;      /// index into TextFlow::glyphs
    int _current_character;  /// index into TextFlow::character
    /** for cursor up/down movement we must maintain the x position where
    we started so the cursor doesn't 'drift' left or right with the repeated
    quantization to character boundaries. */
    double _x_coordinate;

    iterator(TextFlow const *p, int g, int c)
        : _parent_textflow(p), _current_glyph(g), _current_character(c) {_recalculate_x_coordinate();}
    ~iterator();

    int _paragraphIndexToLineIndex(int paragraph_index) const;
    int _lineIndexToChunkIndex(int line_index) const;
    int _chunkIndexToSpanIndex(int chunk_index) const;
    int _spanIndexToCharacterIndex(int span_index) const;
    int _spanIndexToGlyphIndex(int span_index) const;
};
